/*
 * Created on Jan 30, 2004
 */
package messaging;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;

import util.*;

/**
 * Base class for fan out connectors over TCP.
 * Known implementations: edu.cmu.aura.prism.connector.PrismToSupplierConnector,...
 */
public abstract class AbstractFanOutConnector
						implements IConnector
{

	protected Hashtable _remoteEndpoints = new Hashtable();
	protected PooledListener _listener;

	protected String _localHost = null;
	protected int _localPort;

	// for queueing messages
	private 	IMessageQueue		_incomingQueue;
	private 	IMessageQueue		_outgoingQueue;

	// for processing read messages into receive() method
	private 	MessageProcessorThread	_messageProcessor;

	// for sending messages from the outgoing queue
	private		MessageSenderThread     _messageSender;

	/**
	 * Must be called by constructor method(s) in concrete classes
	 * @param localHost
	 * @param localPort
	 * @param dontKnowTheName
	 * @throws java.io.IOException
	 */
	protected final void setLocalEndpoint(String localHost,
											int localPort,
											String dontKnowTheName)
						throws java.io.IOException
	{
		_localHost = localHost;
		_localPort = localPort;
		_incomingQueue = new MessageQueue();
		_outgoingQueue = new MessageQueue();

		_listener = new PooledListener(		_localHost,
											_localPort,
											dontKnowTheName,
											_incomingQueue);

		_messageProcessor = new MessageProcessorThread( dontKnowTheName,
														_incomingQueue,
														this);

		_messageSender = new MessageSenderThread( dontKnowTheName,
												  _outgoingQueue,
												  this);
	}

	/**
	 * Attaches a new remote host to the connector.
	 * Each remote point must be registered before
	 * sending any message. Repeated registrations
	 * are refused.

	 * @param remoteHost
	 * @param remotePort
	 * @return true if attachment is successful, false otherwise
	 */
	public final boolean addRemoteEndpoint(
							String remoteHost,
							int remotePort)
	{
		String endpoint = remoteHost + ":" + remotePort;
		ISender sender = (ISender) _remoteEndpoints.get(endpoint);

		if (sender == null)
		{
			sender = new SocketSender(remoteHost, remotePort);
			_remoteEndpoints.put(endpoint, sender);
		}
		else
		{
			sender.addEndPoint();
		}
		return true;
	}

	/**
	 * Removes the attachment to a remote endpoint.
	 * Does nothing if the endpoint was not
	 * previously registered.
	 * @param remoteHost
	 * @param remotePort
	 */
	public final void removeRemoteEndpoint(	String remoteHost,
											int remotePort)
	{
		String endpoint = remoteHost + ":" + remotePort;
		ISender sender = (ISender) _remoteEndpoints.get(endpoint);

		if (sender != null)
		{
			if ( sender.removeEndPoint() == 0 )
				_remoteEndpoints.remove(endpoint);
		}
	}

	/**
	 * @return Host address of local endpoint
	 */
	public final String getLocalHost()
	{
		return _localHost;
	}

	/**
	 * @return Port number of local endpoint
	 */
	public final int getLocalPort()
	{
		return _localPort;
	}

	/**
	 * @return Enumeration of the remote endpoints in the form (String) host:port
	 */
	public final Enumeration getRemoteEndPoints()
	{
		return _remoteEndpoints.keys();
	}

	/**
	 * Queues sending a message to the remote endpoint. This method is non-blocking
	 * and may result in a subsequent call to errorSending().
	 * The remote endpoint must have previously be attached.
	 * @param message
	 * @param remoteHost
	 * @param remotePort
	 * @return true if the endpoint is currently attached, false otherwise
	 * @see edu.cmu.util.AbstractFanOutConnector#errorSending(String, String, int, IOException)
	 */
	public final boolean send(		String message,
									String remoteHost,
									int remotePort)
	{
		String endpoint = remoteHost + ":" + remotePort;

		if (_remoteEndpoints.containsKey(endpoint))
		{
			try
			{
				MessageWrapper wrapper = new MessageWrapper();
				wrapper._message = message;
				wrapper._sender = (ISender) _remoteEndpoints.get(endpoint);
				_outgoingQueue.queueMessage(wrapper);
			}
			catch ( java.lang.InterruptedException interrupted)
			{
				Assert.report(interrupted);
			}
			return true;
		}
		else
			return false;
	}

	/**
	 * Called if a previous send attempt fails.
	 * Concrete classes must implement this method to react to the send failure appropriately.
	 * @param message that couldn't be sent
	 * @param exception
	 */
	public abstract void errorSending(
		String message,
		String remoteHost,
		int remotePort,
		IOException exception);

	/**
	 * Called after the connector receives a message from a remote host.
	 * Concrete classes must implement this method to parse the message and dispatch it appropriately
	 * @param message received from remote host
	 */
	public abstract void receive(String message);
}
