/*
  File: PooledListener.java

  History:
  Date       	Who				What
  May_19_2003  	poladian		Created public version
*/

package messaging;
import java.net.*;
import java.util.*;

import util.AuraSleep;

import EDU.oswego.cs.dl.util.concurrent.*;



/**
 *  This listener internally makes use of a thread pool in order
 *  to manage the reading of messages from sockets.
 *
 *  This listener derives from class java.lang.Thread, and accepts
 *  incoming sockets in its thread.  Once accepted, sockets are
 *  placed into an internally managed unbounded queue as job objects.
 *
 *  There is a dedicated thread, called the QueueWatcher, which
 *  periodically takes jobs off the socket job queue, and schedules
 *  them for execution using a PooledExecutor
 *  (see  EDU.oswego.cs.dl.util.concurrent.PooledExecutor)
 *
 *  The processing of incoming messages in sockets is done using
 *  edu.cmu.aura.connector.YieldingSocketProcessor objects.
 *
 *  These are called yielding, because they processes up to a
 *  prescribed amount of messages (currently 5), and then yield
 *  the processing to other jobs  (but they are placed back into
 *  the accepted sockets kobs queue for later processing.
 *
 *  @see EDU.oswego.cs.dl.util.concurrent.PooledExecutor
 *  @see edu.cmu.util.YieldingSocketReader
 **/

class PooledListener extends Thread
{

	/**
	 *  This internal thread manages the queue of ready sockets.
	 *  It takes a job (with a socket inside) off the queue, and
	 *  then inserts into the thread pool.  The job wrapper for
	 *  the sockets is provided by YieldingSocketProcessor class.
	 *
	 *  To ensure it is not spinning cycles, it has a fixed back-off
	 *  of 100 milliseconds after 100 attempts.
	 *
	 *  This thread will permanently dump a job off the queue, if
	 *  that job becomes dumpable as defined by the underlying socket.
	 *
	**/

	class QueueWatcher extends Thread
	{
		/* this is a queue of sockets waiting to be processed */

		protected Channel			m_queueToWatch;
		protected Executor			m_pooledExecToOffload;

		QueueWatcher(Channel channel,
							Executor queue)
		{
			super("PooledListener-QueueWatcher");
			if ( channel == null || queue == null )
				throw new IllegalArgumentException();

			m_pooledExecToOffload = queue;
			m_queueToWatch = channel;
		}

		public void run()
		{
//			System.out.println(" @@@ QueueWatcher :: starting.");
			long count = 0;
			long jobsDumped = 0;
			long idleSpin = 0;
			boolean bWasInterrupted = false;

			while ( !m_bStopRunning )
			{
				for ( int i =0; i <100; i ++ )
				{
					count ++;
					YieldingSocketReader job = null;
					try
					{
						job = (YieldingSocketReader)
												m_queueToWatch.take();
						if ( job.canDump() )
						{
							System.out.println("@@@@ Dumping job " + job.getId());
							job = null;
							jobsDumped ++;
							continue;
						}

						if ( ! job.isScheduled() )
						{
							job.schedule();
								m_pooledExecToOffload.execute(job);
						}
						else
							idleSpin++;
					}
					catch (InterruptedException intExc)
					{
						/* xxx */
						intExc.printStackTrace();
						bWasInterrupted = true;
					}
					if ( job != null )
					{
						try
						{
							m_queueToWatch.put(job);
						}
						catch (InterruptedException iExc)
						{
							/* xxx */
							job.prepareToDump();
							iExc.printStackTrace();
							bWasInterrupted = true;
						}
					}

					if ( bWasInterrupted)
						Thread.currentThread().interrupt();
				}

				if ( bWasInterrupted )
					break;

//				System.out.println("@@@@ Socket Queue Watcher status.");
				AuraSleep.sleep(100);

/*				if ( (count % 1000) == 0 )
				{
					System.out.println(	" #### count " + count +
						", dumped " + jobsDumped +	", idle spin " + idleSpin);

				}*/
			}

			if ( m_bStopRunning || bWasInterrupted )
			{
				int i = 0;
				System.out.println(" @@@ QueueWatcher :: interrupt signal received.  Draining jobs.");
				while ( true )
				{
					YieldingSocketReader job = null;
					try
					{
						job = (YieldingSocketReader)
											m_queueToWatch.take();
					}
					catch ( java.lang.InterruptedException intExc)
					{
						intExc.printStackTrace();
						break;
					}
					if ( job != null )
					{
						job.prepareToDump();
						i++;
					}
				}
				System.out.println(" @@@ QueueWatcher :: interrupt signal received.  Finished draining : " + i) ;
//				System.out.println(" @@@ QueueWatcher :: interrupt signal received.  Aborting.");
			}
			System.out.println(" @@@ QueueWatcher :: interrupt signal received.  Ending.");
		}
	}

	class NamedThreadFactory
		implements EDU.oswego.cs.dl.util.concurrent.ThreadFactory
	{
		protected Integer index = new Integer(1);
		public Thread newThread(java.lang.Runnable runnable)
		{
			synchronized (this)
			{
				int i = index.intValue();
				index = new Integer(i+1);
				return new Thread(runnable, m_strName + "-Pool-" + i);
			}
		}
	}



	/* these are for the internal management of the listener */

	/**
	*  We put into this data structure all the accepted sockets.
	*  This queue is watched by the QueueWatcher thread
	*  which perdiodically takes items (jos) off this queue
	*  and puts into the Pooled Executor.
	**/
	protected LinkedQueue			m_acceptedSocketsQueue;

	/**  This is a special thread that perdiocally takes items
	*	off the above queue and puts them into the executor.
	**/

	protected QueueWatcher			m_socketQueueWatcher;

	/**  This is a pooled executor, with an internal configurable-sized
	*	buffer, and a configurable pool of threads.  Jobs placed on
	*   it are first put on the buffer, and then consumed by threads
	*   read the documentation of the class for more detail.
	**/

	protected PooledExecutor		m_pooledExecutor;

	/**  This is the server socket.   The main thread blocks on the accept() method of this object.
	*/
	protected ServerSocket 			m_serverSocket;

	/**  This is a synchronized queue, onto which messages are placed for later processing.
	*  This queue is provided by the user of this class, via a constructor.
	*	The interface of this queue is provided by: IMessageQueue
	*	@see edu.cmu.aura.connector.IMessageQueue
	**/

	protected IMessageQueue		m_MessageQueue;

	protected boolean				m_bStopRunning = false;

	protected String 				m_strName ;

	/**
	*   This is a set of all incoming sockets.
	*/
	protected Set					m_allSockets;



	/**
	*   this constructor takes a local listen port and a IMessageQueue object
	*   configured in this manner, this listener will listen on all local
	*   network interfaces at the specified port.  if a specific network
	*   address is desired, then another constructor should be used.
	*   start() method is called at the end of the constructor, resulting
	*   in the spawing of this thread, and the calling of the run() method
	*
	**/
	protected PooledListener (	int port,
								IMessageQueue queue,
								String strName)
							throws java.io.IOException
	{
		m_serverSocket = new ServerSocket(port);
		m_MessageQueue = queue;
		m_strName = strName;
		initializeMembers();
		this.start();
	}


	/**
	*   this constructor takes a local listen port, address, and a IMessageQueue object
	*   start() method is called at the end of the constructor, resulting
	*   in the spawing of this thread, and the calling of the run() method
	**/
	PooledListener (	String strAddress,
							int port,
							String strName,
							IMessageQueue queue)
							throws java.io.IOException
	{
		InetAddress [] addresses = InetAddress.getAllByName(strAddress);
		if ( addresses != null && addresses.length >0 )
		{
			m_serverSocket = new ServerSocket(  port,
												0,
												addresses[0]);
		}
		else
			m_serverSocket = new ServerSocket(port);

		m_strName = strName;
		m_MessageQueue = queue;
		initializeMembers();
		this.start();
	}


	/**
	*   called by all constructors.  initializes the internal data structurers.
	*   creates a pooled executor ( a thread pool), with a Channel of type
	*   BoundedBuffer.  The size of the Buffer is 3.  The maximum pool size is 50,
	*   and the minium pool size is 20.  Keep alive parameter is 30 seconds.
	*   The pool is started with 20 threads.  For exaplanation on how the pools
	*   behavior is affected by this parameters, read the javadocs for PooledExecutor
    *
	*  @see EDU.oswego.cs.dl.util.concurrent.PooledExecutor
	*  @see EDU.oswego.cs.dl.util.concurrent.Channel
	*  @see EDU.oswego.cs.dl.util.concurrent.BoundedBuffer
	**/
	protected void initializeMembers()
	{
		m_acceptedSocketsQueue = new LinkedQueue();

		m_allSockets = new LinkedHashSet();


		m_pooledExecutor = new PooledExecutor(new BoundedBuffer(3), 25);
		m_pooledExecutor.setThreadFactory(new NamedThreadFactory());
		m_pooledExecutor.setMinimumPoolSize(20);
		m_pooledExecutor.setKeepAliveTime(1000 * 30 );
		m_pooledExecutor.waitWhenBlocked();
		m_pooledExecutor.createThreads(10);

		m_socketQueueWatcher = new QueueWatcher(
									m_acceptedSocketsQueue,
									m_pooledExecutor);

	}

	protected void shutDownSocketServerThroughSideChannel()
	{
		try
		{
			m_serverSocket.close();
		}
		catch ( java.io.IOException ioExc)
		{

		}
	}


	/**
	*   override's the interrupt() method in java.lang.Thread
	*   stops accepting sockets,
	*   closes currently open client sockets,
	*   propagates the interrupt to the underlying thread pool via the shutdownNow() method,
	*   and the QueueWatcher thread
	*
	**/

	public void interrupt()
	{
		System.out.println("@@@@ PooledListener :: interrupt signal received, shutting down.");
		m_bStopRunning = true;
		shutDownSocketServerThroughSideChannel();
		Iterator it = m_allSockets.iterator();

		while (it.hasNext() )
		{
			YieldingSocketReader job = (YieldingSocketReader) it.next();
			job.prepareToDump();
		}

		m_socketQueueWatcher.interrupt();
		m_pooledExecutor.shutdownNow();
		super.interrupt();
	}

	/**
	*   overrides the run() method in java.lang.Thread
	*   starts the QueueWatcher threads.
	*   starts accepting client sockets
	**/

	public void run()
	{
		m_socketQueueWatcher.start();

		System.out.println("@@@@ PooledListener *" + m_strName + "* starting at address "
				+ m_serverSocket.getInetAddress().getHostAddress()
				+ ":" + m_serverSocket.getLocalPort());

		while ( !m_bStopRunning  )
		{
			Socket client = null;
			try
			{

//				System.out.println("@@@@ PooledListener accepting sockets.");

				client = m_serverSocket.accept();
				System.out.println("@@@@ PooledListener accepted socket from "
						+ client.getInetAddress().getHostAddress()
						+ ":" + client.getPort());


				YieldingSocketReader socProc =
							new YieldingSocketReader (
										m_strName,
										client,
										this.m_MessageQueue);

				m_allSockets.add(socProc);
				try
				{
					m_acceptedSocketsQueue.put(socProc);


				}
				catch (InterruptedException intExc)
				{
					/* XXX */
					intExc.printStackTrace();
				}

			}
			catch (java.io.IOException ioExc)
			{
				ioExc.printStackTrace();
			}

		}
//		System.out.println("@@@@ PooledListener: exiting run loop.");
	}

  /**
   *  Returns the listening address of this listener, in a url format.
   *  Typically, this is a string of the form: machine:port
   *  @return  listening address
   **/
	String getListenerAddress()
	{

		String strHost = m_serverSocket.getInetAddress().getHostAddress();
		int iPort = m_serverSocket.getLocalPort();
		String retVal = new String(strHost + ":" + iPort);
		return retVal;
	}
}
